# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 21.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 5, task 1.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation
# pylint: disable=redefined-outer-name
# pylint: disable=import-error
# pylint: disable=singleton-comparison

import numpy as np
import matplotlib.pyplot as plt
import cv2
from skimage import morphology
from scipy.ndimage.morphology import binary_fill_holes


def plot_image(image, fig_num, name=None, colormap='gray', write_file=None):
  """Plot image"""
  fig = plt.figure(fig_num)
  plt.imshow(image, cmap=colormap, interpolation='none')
  if name is not None:
    plt.title(name)
  plt.xticks([]), plt.yticks([])
  plt.tight_layout()
  if write_file:
    fig.savefig(write_file, bbox_inches='tight', pad_inches=0)
  fig_num += 1
  return fig_num

def main():
  """Connected components"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 5')
  print('Task 1')
  print('-'*80)

  # Configurations
  image_file = '../../images/pillsetc.png'
  fig_num = 0

  # Step 1: Read image
  rgb_image = cv2.cvtColor(cv2.imread(image_file, cv2.IMREAD_COLOR), cv2.COLOR_BGR2RGB)
  fig_num = plot_image(rgb_image, fig_num,
                       write_file='results/task01/original_rgb.png')

  # Step 2: Threshold the image
  gray_image = cv2.cvtColor(rgb_image, cv2.COLOR_RGB2GRAY)
  threshold, thresh_image = cv2.threshold(gray_image, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)
  print('Graylevel threshold computed with otsu: ', threshold)
  fig_num = plot_image(thresh_image, fig_num,
                       write_file='results/task01/thresholded.png')

  # Step 3: Remove the noise

  # Remove objects containing fewer than 30 pixels (both gives essentially the
  # same output)
  rmnoise_image = morphology.remove_small_objects(thresh_image.astype(bool), min_size=30)
  #rmnoise_image = morphology.binary_opening(thresh_image)
  fig_num = plot_image(rmnoise_image, fig_num,
                       write_file='results/task01/rmnoise_image.png')

  # Fill a gap in the cap of the pen
  fillcap_image = morphology.binary_closing(rmnoise_image, selem=morphology.disk(2))
  fig_num = plot_image(fillcap_image, fig_num,
                       write_file='results/task01/fillcap_image.png')

  # Fill holes
  clean_image = binary_fill_holes(fillcap_image)
  fig_num = plot_image(clean_image, fig_num,
                       write_file='results/task01/clean_image.png')

  # Step 4: Find boundaries
  # We will only search for external boundaries, that is, no inner boundaries,
  # or boundaries in enclaves (making the filling of the holes in the previous
  # step redundant, but nevermind). We also chose to find all contour points.
  # See more at (but note that these links are for opencv2)
  # http://docs.opencv.org/2.4/modules/imgproc/doc/structural_analysis_and_shape_descriptors.html
  # http://docs.opencv.org/3.0-beta/doc/py_tutorials/py_imgproc/py_contours/py_contours_begin/py_contours_begin.html
  #
  # For opencv3, see more at
  # http://www.pyimagesearch.com/2015/08/10/checking-your-opencv-version-using-python/
  # http://www.pyimagesearch.com/2016/02/01/opencv-center-of-contour/
  #
  # The most visible change is that in version 2 it is
  # contours, hierarchy = cv2.findContours(...)
  # contrary to version 3 which is
  # label_image, contours,  hierarchy = cv2.findContours(...)

  _, contours, _ = cv2.findContours(clean_image.astype(np.uint8), # pylint: disable=unused-variable
                                    cv2.RETR_EXTERNAL,
                                    cv2.CHAIN_APPROX_NONE)

  # Draw the contours and the objects they enclose
  obj_image = np.zeros(clean_image.shape)
  for ind, cnt in enumerate(contours):
    col = 255 * (ind + 1) / len(contours)
    cv2.drawContours(obj_image, [cnt], 0, col, -1)
  fig_num = plot_image(obj_image.astype(np.uint8), fig_num, colormap='magma',
                       write_file='results/task01/object_image.png')

  # Step 5: Determine which objects are round
  # For a circle, the following relationship will hold
  #
  #         4 * pi * Area / Circumference^2 = 1
  #
  # Therefore, we use the metric alpha = 4*pi*Area/Perimeter^2 to determine how round the
  # shape is, the closer it is to one, the rounder it is. Mark the center of all
  # objects with alpha > round_thresh
  round_thresh = 0.9

  fig = plt.figure(fig_num)
  height, width = obj_image.shape
  plt.imshow(obj_image.astype(np.uint8), cmap='magma', interpolation='none')
  plt.xticks([]), plt.yticks([])

  print('Object      Area Perimeter Roundness')
  for ind, contour in enumerate(contours):
    # Compute a simple estimate of the object perimeter
    # contour is a list of N coordinate points (x_i, y_i), and the perimeter is computed as
    # perimeter = sum_{i=0}^{N-2} ( sqrt( (x_{i+1} - x_i)^2 + (y_{i+1} - y_i)^2 ))
    delta_squared = np.diff(contour, axis=0)*np.diff(contour, axis=0)
    perimeter = np.sum(np.sqrt(np.sum(delta_squared, axis=1)))

    # or use the built in arcLength() (both arcLength() and contourArea() are
    # based on Freeman chain codes).
    perimeter = cv2.arcLength(contour, True)

    area = cv2.contourArea(contour)
    alpha = 4*np.pi*area/(perimeter**2)
    print('{0:>6} {1:>9,.0f} {2:>9,.2f} {3:>9,.3f}'.format(ind, area, perimeter, alpha))

    # Compute moments in order to find the centroid of objects
    moments = cv2.moments(contour)
    cx = int(moments['m10'] / moments['m00'])
    cy = int(moments['m01'] / moments['m00'])

    # Plot labels
    plt.text(cx, cy, str(ind), color='green', fontsize=14, weight='bold')

    # Plot a red circle on the centroid of the round objects
    if alpha > round_thresh:
      plt.plot(cx, cy, 'ro')

  # Fit the plt.plot() figures to the plt.imshow()
  axes = plt.gca()
  axes.set_xlim([0, width-1])
  axes.set_ylim([height-1, 0])
  plt.tight_layout()
  fig.savefig('results/task01/marked_image.png', bbox_inches='tight', pad_inches=0)
  fig_num += 1


  plt.show()

if __name__ == "__main__":
  main()
